import Control.Monad hiding (join)
import Data.Array.Unboxed
import System.Environment

import Control.Parallel.Strategies
import Control.DeepSeq
import Future

import System.Time
import Text.Printf

type FpType = Float


data ParameterSet = ParameterSet {
        sptprice :: FpType,
        strike :: FpType,
        rate :: FpType,
        volatility :: FpType ,
        otime :: FpType,
        otype :: Bool
} deriving Show


data_init :: Array Int ParameterSet

-- This defines some hard coded data as a big constant array:
#include "blackscholes_data.hs"
size_init = let (s,e) = bounds data_init in e - s + 1

inv_sqrt_2xPI = 0.39894228040143270286

cndf :: FpType -> FpType
cndf inputX = if sign then 1.0 - xLocal else xLocal
  where
    sign = inputX < 0.0
    inputX' = if sign then -inputX else inputX
    
    -- Compute NPrimeX term common to both four & six decimal accuracy calcs
    xNPrimeofX = inv_sqrt_2xPI * exp(-0.5 * inputX * inputX);

    xK2 = 1.0 / (0.2316419 * inputX + 1.0);
    xK2_2 = xK2 * xK2; -- Need all powers of xK2 from ^1 to ^5:
    xK2_3 = xK2_2 * xK2;
    xK2_4 = xK2_3 * xK2;
    xK2_5 = xK2_4 * xK2;
    
    xLocal = 1.0 - xLocal_1 * xNPrimeofX;
    xLocal_1 = xK2 * 0.319381530 + xLocal_2;
    xLocal_2 = xK2_2 * (-0.356563782) + xLocal_3 + xLocal_3' + xLocal_3'';
    xLocal_3 = xK2_3 * 1.781477937;
    xLocal_3' = xK2_4 * (-1.821255978);
    xLocal_3'' = xK2_5 * 1.330274429;


blkSchlsEqEuroNoDiv :: FpType -> FpType -> FpType -> FpType -> FpType -> Bool -> Float -> FpType
blkSchlsEqEuroNoDiv sptprice strike rate volatility time otype timet =
   if not otype
   then (sptprice * nofXd1) - (futureValueX * nofXd2)
   else let negNofXd1 = 1.0 - nofXd1
            negNofXd2 = 1.0 - nofXd2
        in (futureValueX * negNofXd2) - (sptprice * negNofXd1)
 where
   logValues = log( sptprice / strike )
   xPowerTerm = 0.5 * volatility * volatility
   xDen = volatility * sqrt(time)
   xD1 = (((rate + xPowerTerm) * time) + logValues) / xDen
   xD2 = xD1 - xDen

   nofXd1 = cndf xD1
   nofXd2 = cndf xD1
   futureValueX = strike * exp ( -(rate) * (time) )

executeStep :: Int -> Int -> UArray Int FpType
executeStep t granularity =
       listArray (0, granularity-1) $
                 Prelude.map (\i ->
                              let ParameterSet { .. } = data_init ! ((t+i) `mod` size_init)
                         in blkSchlsEqEuroNoDiv sptprice strike rate volatility otime otype 0)
                 [0 .. granularity-1]
                
blackscholes :: Int -> Int -> Eval Float
blackscholes numOptions granularity =
   do
      fs <- forM [0, granularity .. numOptions-1] $ \t ->
              fork (return (executeStep t granularity))

      foldM (\ acc f ->
                 do x <- join f
                    return (acc + (x ! 0)))
				0 fs

main = do args <- getArgs 
          let (numOptions, granularity) =
               case args of 
  	         []      -> (10000, 1000)
  	         [b]     -> (10, read b)
	         [b,n] -> (read n, read b) -- number of options and granularity

          if granularity > numOptions
	   then error "Granularity must be bigger than numOptions!!"
	   else return ()
                
          t1 <- getClockTime
	  putStrLn$ "Running blackscholes, numOptions "++ show numOptions ++ " and block size " ++ show granularity
          let result = runEval $ blackscholes numOptions granularity
	  putStrLn$ "Final result: "++ show result
	  return result
          t2 <- getClockTime
          putStrLn $ printf "time taken: %.2fs" $ secDiff t1 t2


-- func to calc time taken between t0 and t1 in sec
secDiff::ClockTime -> ClockTime -> Float
secDiff (TOD secs1 psecs1) (TOD secs2 psecs2) = fromInteger (psecs2 - psecs1) / 1e12 + fromInteger (secs2 - secs1)




			

